/*-----------------------------------------------------------------------

File  : code_generator.c

Author: Stephan Schulz

Contents

  Actual code for the code generator 

  Copyright 1998-2002 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Wed Oct 30 18:20:31 EST 2002
    New

-----------------------------------------------------------------------*/

#include "code_generator.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

char *static_preamble =
"#include <stdlib.h>\n"
"#include <stdio.h>\n"
"\n"
"#define INC(X) (X)++\n"
"#define DEC(X) (X)--\n"
"#define ASSIGN(X,Y) (X) = (Y)\n"
"#define INIT0(X) (X)=0\n"
"#define ZERO(X) ((X)==0)\n"
"#define INITFROMSTRING(X,S) (X) = atol(S)\n"
"#define PRINTPLINT(X) printf(\"%u\",X)\n"
"\n"
"typedef int PLINT;\n"
"\n"
"int main(int argc, char *argv[])\n"
"{\n";

char *static_postamble = "   return EXIT_SUCCESS;\n}\n";


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/




/*-----------------------------------------------------------------------
//
// Function: PrintStatementC()
// 
//   Print a PL statement in internal encoding back as equivalent C
//   code (assuming suitable macro definitions);
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void PrintStatementC(Statement_p st, int indent)
{
   if(st->label)
   {
      printf("%-4s: ", st->label);      
   }
   else
   {
      printf("      ");      
   }
   PrintIndent(indent);
   switch(st->type)
   {
   case StInc:
	 printf("INC(%s);\n",st->argument);
	 break;
   case StNull:
	 printf("INIT0(%s);\n",st->argument);
	 break;
   case StGoto:
	 printf("goto %s;\n",st->argument);
	 break;
   case StLoop:
	 printf("{\n");
	 PrintIndent(indent);
	 printf("         PLINT loopvar%ld;\n", st->statement_id);
	 PrintIndent(indent);
	 printf("         for(ASSIGN(loopvar%ld,%s);!ZERO(loopvar%ld); DEC(loopvar%ld))\n",
		st->statement_id, st->argument, st->statement_id, st->statement_id);
	 PrintIndent(indent);
	 printf("         {\n");
	 PrintProgramC(st->subprog, indent+2);
	 PrintIndent(indent);
	 printf("         }\n");
	 PrintIndent(indent); 
	 printf("      }\n");
	 break;
   default:
	 Error("Unknown statement type in statement");
	 break;
   }
}


/*-----------------------------------------------------------------------
//
// Function: PrintProgramC()
//
//   Print a PL program back as C code
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void PrintProgramC(Prog_p prog, int indent)
{
   if(prog)
   {
      PrintStatementC(prog->statement, indent);
      PrintProgramC(prog->prog_rest, indent);
   }
}



/*-----------------------------------------------------------------------
//
// Function: CollectProgVars()
//
//   Collect all the variables in prog, sorted by initial letter.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void CollectProgVars(Prog_p prog, BinTree_p *X, BinTree_p *Y, 
                     BinTree_p *Z)
{
   if(prog)
   {
      CollectStatementVars(prog->statement, X,Y,Z);
      CollectProgVars(prog->prog_rest, X,Y,Z);
   }
}


/*-----------------------------------------------------------------------
//
// Function: CollectStatementVars()
//
//   Collect variables in statement (and substatements)
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void CollectStatementVars(Statement_p statement, BinTree_p *X,
			  BinTree_p *Y, BinTree_p *Z)
{
   if(statement->type != StGoto)
   {
      switch(statement->argument[0])
      {
      case 'X':
	    BinTreeInsert(X, atoi(&statement->argument[1]));
	    break;
      case 'Y':
	    BinTreeInsert(Y, atoi(&statement->argument[1]));
	    break;
      case 'Z':
	    BinTreeInsert(Z, atoi(&statement->argument[1]));
	    break;	    
      }
   }
   if(statement->type == StLoop)
   {
      CollectProgVars(statement->subprog, X,Y,Z);
   }
}


/*-----------------------------------------------------------------------
//
// Function: PrintVarDefs()
//
//   Print definitions for variables stored in tree.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void PrintVarDefs(BinTree_p tree, char* basename)
{
   if(tree)
   {
      PrintVarDefs(tree->lchild, basename);
      printf("   PLINT %s%d;\n",basename,tree->key);
      PrintVarDefs(tree->rchild, basename);   
   }
}



/*-----------------------------------------------------------------------
//
// Function: PrintVarInits()
//
//   Print initialization code for variables.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void PrintVarInits(BinTree_p tree, char* basename, bool input)
{
   if(tree)
   {
      PrintVarInits(tree->lchild, basename, input);
      if(input)
      {
	 printf("   if(argc>%d)\n",tree->key);
	 printf("   {\n");
	 printf("      INITFROMSTRING(%s%d,argv[%d]);\n",basename,tree->key,tree->key);
         printf("   }\n");
         printf("   else\n");
	 printf("   {\n");
	 printf("      INIT0(%s%d);\n",basename,tree->key);
         printf("   }\n"); 
      }
      else
      {
	 printf("   INIT0(%s%d);\n",basename,tree->key);
      }
      PrintVarInits(tree->rchild, basename,input);   
   }
}



/*-----------------------------------------------------------------------
//
// Function: PrintVarValues()
//
//   Print print instructions for variables in tree.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void PrintVarValues(BinTree_p tree, char* basename)
{
   if(tree)
   {
      PrintVarValues(tree->lchild, basename);
      printf("   printf(\"%s%d = \");\n",basename,tree->key);
      printf("   PRINTPLINT(%s%d);\n",basename,tree->key);
      printf("   printf(\"\\n\");\n");
      PrintVarValues(tree->rchild, basename);   
   }
}



/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/






